from unittest.mock import Mock, patch

from eventsourcing_helpers.event_handler import EventHandler

module = 'eventsourcing_helpers.event_handler'


class EventHandlerTests:
    def setup_method(self):
        self.event_class, self.id = 'FooEvent', 1
        self.message = Mock(
            value={
                'class': self.event_class,
                'data': {
                    'id': self.id
                }
            }
        )
        self.event = Mock()
        self.event._class = self.event_class
        self.event.id = self.id

        self.message_deserializer = Mock()
        self.message_deserializer.return_value = self.event

        self.event_handler = Mock()
        self.handler = EventHandler
        self.handler.handlers = {self.event_class: self.event_handler}
        self.handler = self.handler(self.message_deserializer)

    @patch(f'{module}.EventHandler._handle_event')
    @patch(f'{module}.EventHandler._can_handle_command')
    def test_handle(self, mock_can_handle, mock_handle):
        """
        Test that correct methods are invoked when handling an event.
        """
        self.handler.handle(self.message)

        mock_can_handle.assert_called_once_with(self.message)
        mock_handle.assert_called_once_with(self.event)

    def test_handle_event(self):
        """
        Test that the correct event handler is invoked.
        """
        self.handler._handle_event(self.event)
        self.event_handler.called_once_with(self.event)

    def test_can_handle_command(self):
        """
        Test that we only handle registered events.
        """
        can_handle = self.handler._can_handle_command(self.message)
        assert can_handle is True

        self.message.value['class'] = 'BarEvent'
        can_handle = self.handler._can_handle_command(self.message)
        assert can_handle is False



import json
import re
import uuid
from itertools import chain
from typing import Any, Callable, Iterator, List, Union

import jsonpickle
import structlog

from eventsourcing_helpers.utils import get_all_nested_keys

word_regexp = re.compile('[A-Z][a-z]+|[A-Z]+(?![a-z])')
logger = structlog.get_logger(__name__)


class MissingEntityApplyMethod(Exception):
    pass


class Entity:
    """
    A rich domain model that exposes attributes and behaviour
    with an identity and a life cycle.
    """
    # a shared list between all entities with staged events
    # that later will be committed to the repository.
    _events: List[Any] = []

    def __init__(self) -> None:
        self.id: Union[str, None] = None
        self._version: int = 0

    def __call__(self, *args, **kwargs):
        # fixes https://github.com/python/mypy/issues/2113
        super().__call__(*args, **kwargs)

    def __repr__(self) -> str:
        attrs = {k: v for k, v in self.__dict__.items() if v is not None}
        return f"{self._class}({attrs})"

    def get_representation(self) -> str:
        json_representation = jsonpickle.encode(self)
        dict_representation = json.loads(json_representation)
        keys = get_all_nested_keys(dict_representation, [])

        return self.__class__.__name__ + ', ' + ', '.join(keys)

    @property
    def _class(self):
        return self.__class__.__name__

    def _get_apply_method(self, entity: Any, method_name: str) -> Callable:
        """
        Get apply method on an entity.

        Args:
            entity: Entity to look on.
            method_name: Apply method name to look for.

        Returns:
            method: Callback to apply method.
        """
        return getattr(entity, method_name)

    def _apply_event(
        self, event: Any, entity: 'Entity', method_name, is_new
    ) -> None:
        """
        Apply an event on one entity.

        Args:
            event: Event to be applied.
            entity: Entity to apply the event on.
            method_name: Apply method on the entity.
            is_new: Flag to indicate if the event should be staged for commit.
        """
        try:
            apply_method = self._get_apply_method(entity, method_name)
            # TODO: apply the event in the aggregate root if it's defined.
        except AttributeError:
            raise MissingEntityApplyMethod(f"{entity._class}.{method_name}")
        else:
            if is_new:
                logger.debug(
                    "Applying event", method=method_name, id=event.id,
                    event_class=event._class, entity_class=entity._class
                )
            apply_method(event)
            self._stage_event(event, is_new)

    def _stage_event(self, event: Any, is_new: bool) -> None:
        """
        Stage an event.

        All staged events in the list will later be committed to
        the repository.

        Args:
            event: Event to be staged.
            is_new: Flag that indicates if the event should be staged.
        """
        if is_new:
            logger.info("Staging event", event_class=event._class)
            self._events.append(event)

    def _get_child_entities(self) -> Iterator['Entity']:
        """
        Get all child entities for the current instance
        including all instances from all EntityDict's.

        Returns:
            iterable: A list with all child entities.
        """
        entities = [
            e._get_all_entities()
            for e in self.__dict__.values()
            if isinstance(e, (Entity, EntityDict))
        ]
        return chain.from_iterable(entities)

    def _get_all_entities(self) -> Iterator['Entity']:
        """
        Get the current instance and all child entities.

        Returns:
            iterable: A list with all entities.
        """
        return chain([self], self._get_child_entities())

    def _get_entity(self, id: str) -> 'Entity':
        """
        Find and return an entity instance with the given id.

        If no entity are found we return the current instance.
        This is a normal situation when an child entity has not yet
        been created by the parent.

        TODO: This will probably not work if we have more then two
        levels of entities. We should always return the child entities
        parent, not aggregate root.

        Args:
            id: The id of the entity.

        Returns:
            Entity: Found entity or current instance.
        """
        entities = self._get_all_entities()
        return next((e for e in entities if e.id == id), self)

    def _get_apply_method_name(self, event_class: str) -> str:
        """
        Gets the apply method name for a given event.

        Args:
            event_class: Name of the event. Probably the class
                name of the event.

        Returns:
            str: Name of the apply method.

        Example:
            >>> event = FooEvent(id=1)
            >>> Entity._get_apply_method_name(event._class)
            'apply_foo_event'
        """
        words = re.findall(word_regexp, event_class)
        lowered_words = list(map(str.lower, words))
        apply_method_name = 'apply_' + '_'.join(lowered_words)

        return apply_method_name

    def _clear_staged_events(self) -> None:
        """
        Clear staged events from ALL Entity instances.
        """
        logger.info("Clearing staged events")
        Entity._events = []

    def _apply_events(self, events: List[Any],
                      ignore_missing_apply_methods: bool = False) -> None:
        """
        Apply multiple events.

        Args:
            events: A list of events.
        """
        logger.info("Apply events from repository")
        for event in events:
            try:
                self.apply_event(event, is_new=False)
            except MissingEntityApplyMethod:
                if not ignore_missing_apply_methods:
                    raise

    def create_id(self) -> str:
        """
        Returns an unique id as a string.
        """
        return str(uuid.uuid4())

    def apply_event(self, event: Any, is_new: bool = True) -> None:
        """
        Applies an event by calling the correct entity apply method.

        The apply methods performs all state changes in the aggregate.

        Args:
            event: Event to be applied.
            is_new: Flag that indicates if the event should be staged
                for commit.
        """
        apply_method_name = self._get_apply_method_name(event._class)
        entity = self._get_entity(event.id)

        self._apply_event(event, entity, apply_method_name, is_new)


class EntityDict(dict):
    """
    A collection of domain entities implemented as a dict to allow
    fast lookup by a key.
    """

    def __repr__(self) -> str:
        return f"{self._class}({self.values()})"

    def __setitem__(self, key: str, value: Entity) -> None:
        assert isinstance(value, Entity)
        super().__setitem__(key, value)

    @property
    def _class(self):
        return self.__class__.__name__

    def _get_child_entities(self) -> Iterator[Entity]:
        """
        Get all child entities.

        Returns:
            iterable: A list of all child entities.
        """
        entities = [
            e._get_all_entities()
            for e in self.values()
            if isinstance(e, (Entity, EntityDict))
        ]
        return chain.from_iterable(entities)

    def _get_all_entities(self) -> Iterator[Entity]:
        """
        Get all entities.

        This method only exist so the call in `_get_child_entities`
        methods (Entity, EntityDict) does not break.

        Returns:
            iterable: A list of all child entities.
        """
        return self._get_child_entities()


class AggregateRoot(Entity):
    """
    The aggregate root represents a single entity and acts as a gateway for all
    modifications within the aggregate.

    The aggregate root may or may not contain an object graph which represents
    a logical cohesive group of domain models (entities, value objects).

    The aggregate root is meant to be used as a facade which means that
    all commands must go through it. This gives us the flexibility to run
    business logic in multiple models in the same command.
    """
    pass



from collections import defaultdict
from copy import deepcopy


def get_snapshot_config(config) -> dict:
    config = deepcopy(config)

    snapshot_config = defaultdict(
        dict, config.get('snapshot', {})
    )
    return snapshot_config



from collections import defaultdict
from copy import deepcopy


def get_producer_config(config):
    config = deepcopy(config)
    producer_config = config.get('producer')
    if not producer_config or not producer_config.get('enabled', True):
        return None

    producer_config.update({
        'bootstrap.servers': config.get('bootstrap.servers'),
        'schema.registry.url': config.get('schema.registry.url')
    })
    return producer_config


def get_loader_config(config):
    config = deepcopy(config)
    loader_config = config.get('loader')
    if not loader_config or not loader_config.get('enabled', True):
        return None

    loader_config = defaultdict(dict, loader_config)
    loader_config.update({
        'bootstrap.servers': config.get('bootstrap.servers'),
        'schema.registry.url': config.get('schema.registry.url')
    })
    loader_config['consumer'].update({
        'bootstrap.servers': config.get('bootstrap.servers'),
        'schema.registry.url': config.get('schema.registry.url')
    })
    return loader_config



import copy
from typing import Callable

try:
    from cnamedtuple import namedtuple
except ImportError:
    from collections import namedtuple


class Message:
    """
    Message proxy class.

    Wraps a message class to provide some extra features.

    All attribute lookups are redirected to the wrapped message class.
    """

    def __init__(self, **kwargs) -> None:
        # At this point the instance variable `self._wrapped` is already set.
        #
        # The reason we are saving it in `self.__dict__` is to skip hitting the
        # `__setattr__`dunder method - and thus getting the "Messages are read
        # only" error.
        self.__dict__['_wrapped'] = self._wrapped(**kwargs)  # type: ignore

    @property
    def _class(self) -> str:
        return self._wrapped.__class__.__name__  # type: ignore

    def to_dict(self) -> dict:
        return self._wrapped._asdict()  # type: ignore

    def __eq__(self, other) -> bool:
        return self.__dict__ == other.__dict__

    def __repr__(self) -> str:
        return repr(self._wrapped)  # type: ignore

    def __setattr__(self, *args) -> None:
        raise AttributeError("Messages are read only")

    def __getattr__(self, name: str) -> Callable:
        # redirects all attribute lookups to the real message class.
        raise NotImplementedError


class NewMessage(Message):
    """
    Newly consumed message.

    If we try to access an attribute that doesn't exist we should raise an
    AttributeError.
    """

    def __getattr__(self, name: str) -> Callable:
        attr = getattr(self._wrapped, name)
        return copy.deepcopy(attr)


class OldMessage(Message):
    """
    Message loaded from the repository.

    Accessing attributes that doesn't exist should NOT raise an AtrributeError,
    instead we return None.

    Otherwise we would have to implement try/catch logic in our apply methods
    when we add new fields to our Avro schemas.
    """

    def __getattr__(self, name: str) -> Callable:
        attr = getattr(self._wrapped, name, None)
        return copy.deepcopy(attr)


def message_factory(message_cls: namedtuple, is_new=True) -> type:
    """
    Class decorator used for creating a message proxy class.

    A message proxy is just a message "wrapped" with some extra features and
    where all attribute lookups are passed through to the wrapped message
    class.

    The reason we are doing this is because currently namedtuples doesn't
    support inheritance from a base class - which means we can't easily add
    extra features.

    See: https://github.com/python/typing/issues/427.

    TODO: look into data classes in Python 3.7.

    Args:
        message_cls: Message class to be wrapped.
        is_new: Flag to indicate if the message is new or loaded from the
            repository.

    Returns:
        Message: Message proxy.

    Example:
        >>> @message_factory
        ... class OrderCreated(NamedTuple):
        ...     id: str
        ...     state: str
        >>> event = OrderCreated(id='1', state='open')
        >>> event.id, event.state, event._class
        ('1', 'open', 'OrderCreated')
        >>> event.to_dict()
        {'id': '1', 'state': 'open'}
        >>> isinstance(event, Message)
        True
    """
    # Dynamically create a new class with the same name as the message
    # (event/command) we are wrapping.
    #
    # The new class will inherit from `proxy_cls` and `object` with one
    # instance variable set `_wrapped` which is the actual message
    # (event/command) being wrapped.
    proxy_cls = NewMessage if is_new else OldMessage
    proxy = type(
        message_cls.__name__, (proxy_cls, object), {'_wrapped': message_cls}
    )
    return proxy


# Make the names more explicit for which type of message we are dealing with.
Event = lambda message: message_factory(message)
Command = lambda message: message_factory(message)



from typing import Any

import structlog

from confluent_kafka_helpers.message import Message

from eventsourcing_helpers import metrics
from eventsourcing_helpers.handler import Handler
from eventsourcing_helpers.utils import get_callable_representation

logger = structlog.get_logger(__name__)


class EventHandler(Handler):
    """
    Application service that calls the correct domain handler for an event.
    """

    def _can_handle_command(self, message: Message) -> bool:
        """
        Checks if the event is something we can handle.

        Args:
            message: Consumed message from the bus.

        Returns:
            bool: Flag to indicate if we can handle the event.
        """
        event_class = message.value['class']
        if event_class not in self.handlers:
            logger.debug("Unhandled event", event_class=event_class)
            return False

        return True

    def _handle_event(self, event: Any) -> None:
        """
        Get and call the correct event handler.

        Args:
            event: Event to be handled.
        """
        handler = self.handlers[event._class]

        handler_name = get_callable_representation(handler)
        handler_wrapper = metrics.timed(
            'eventsourcing_helpers.handler.handle',
            tags=['message_type:event', f'message_class:{event._class}', f'handler:{handler_name}']
        )(handler)

        handler_wrapper(event)

    def handle(self, message: dict) -> None:
        """
        Apply correct handler for received event.

        Args:
            message: Consumed message from the bus.
        """
        if not self._can_handle_command(message):
            return

        event = self.message_deserializer(message)
        logger.info("Handling event", event_class=event._class)

        self._handle_event(event)



class MessageBusBackend:
    """
    Message bus interface.
    """

    def produce(self, value: dict, key: str = None, **kwargs) -> None:
        raise NotImplementedError()

    def get_consumer(self):
        raise NotImplementedError()



from typing import Any, Union

import structlog

from confluent_kafka_helpers.message import Message

from eventsourcing_helpers.handler import Handler
from eventsourcing_helpers.metrics import statsd
from eventsourcing_helpers.models import AggregateRoot
from eventsourcing_helpers.repository import Repository
from eventsourcing_helpers.utils import get_callable_representation

logger = structlog.get_logger(__name__)


class CommandHandler(Handler):
    """
    Command handler.

    Handles a command by calling the correct function or method in a handler
    class.
    """

    def _can_handle_command(self, message: Message) -> bool:
        """
        Checks if the command is something we can handle.

        Args:
            message: Consumed message from the bus.

        Returns:
            bool: Flag to indicate if we can handle the command.
        """
        command_class = message.value['class']
        if command_class not in self.handlers:
            logger.debug("Unhandled command", command_class=command_class)
            return False

        return True

    def _handle_command(self, command: Any, handler_inst: Any = None) -> None:
        """
        Get and call the correct command handler.

        The handler can either be a callable or a name of a method in the
        handler instance.

        Args:
            command: Command to be handled.
            handler_inst: Optional handler instance - probably an instance
                of the aggregate root.
        """
        command_class = command._class
        handler = self.handlers[command_class]

        logger.info("Calling command handler", command_class=command_class)
        if handler_inst:
            handler(handler_inst, command)
        else:
            handler(command)

    def handle(self, message: dict) -> None:
        """
        Apply correct handler for the received command.

        Args:
            message: Consumed message from the bus.
        """
        if not self._can_handle_command(message):
            return

        command = self.message_deserializer(message)
        logger.info("Handling command", command_class=command._class)

        command_class = command._class
        handler = self.handlers[command_class]
        handler_name = get_callable_representation(handler)
        with statsd.timed(
            'eventsourcing_helpers.handler.handle', tags=[
                'message_type:command', f'message_class:{command_class}', f'handler:{handler_name}'
            ]
        ):
            self._handle_command(command)


class ESCommandHandler(CommandHandler):
    """
    Event sourced command handler.

    Unlike the command handler we expect an aggregate root to handle all the
    commands.

    The resulting staged events are published to a message bus and persisted in
    an event store using a repository.
    """
    aggregate_root: Union[AggregateRoot, None] = None
    repository_config: Union[dict, None] = None

    def __init__(self, *args, repository: Any = Repository, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert self.aggregate_root
        assert self.repository_config

        self.repository = repository(self.repository_config, self.aggregate_root, **kwargs)

    def _get_aggregate_root(self, id: str) -> AggregateRoot:
        """
        Get latest state of the aggregate root.

        Args:
            id: ID of the aggregate root.

        Returns:
            AggregateRoot: Aggregate root instance with the latest state.
        """
        return self.repository.load(id)

    def _commit_staged_events(self, aggregate_root: AggregateRoot) -> None:
        """
        Commit staged events to the repository.

        Args:
            aggregate_root: Aggregate root with staged events.
        """
        self.repository.commit(aggregate_root)

    def handle(self, message: dict) -> None:
        """
        Apply correct handler for the received command.

        After the command has been handled the staged events are committed to
        the repository.

        Args:
            message: Consumed message from the bus.
        """
        if not self._can_handle_command(message):
            return

        command = self.message_deserializer(message)
        logger.info("Handling command", command_class=command._class)

        command_class = command._class
        handler = self.handlers[command_class]
        handler_name = get_callable_representation(handler)
        with statsd.timed(
            'eventsourcing_helpers.handler.handle', tags=[
                'message_type:command', f'message_class:{command_class}', f'handler:{handler_name}'
            ]
        ):
            aggregate_root = self._get_aggregate_root(command.id)
            try:
                self._handle_command(command, handler_inst=aggregate_root)
            except Exception as e:
                self.repository.snapshot.delete(aggregate_root)
                statsd.increment(
                    'eventsourcing_helpers.snapshot.cache.delete', tags=[f'id={aggregate_root.id}']
                )
                raise e
            self._commit_staged_events(aggregate_root)



from unittest.mock import MagicMock

import pytest

from eventsourcing_helpers.models import AggregateRoot
from eventsourcing_helpers.repository import import_backend


@pytest.fixture(scope='function')
def aggregate_root_cls_mock():
    def aggregate_root_cls(attrs={}, exhaust_events=True, **kwargs):
        mock = MagicMock(spec=AggregateRoot, **kwargs)
        # since the events is a generator we must exhaust it to trigger
        # the call to `backend.load`.
        if exhaust_events:
            apply_events = lambda events, **kwargs: [e for e in events]
        else:
            apply_events = lambda events, **kwargs: events

        default_attrs = {
            'return_value._apply_events.side_effect': apply_events,
            'return_value.get_representation.return_value': 'adsafasf'}
        mock.configure_mock(**{**default_attrs, **attrs})
        return mock

    return aggregate_root_cls


@pytest.fixture(scope='function')
def importer_mock():
    def importer(return_value, **kwargs):
        mock = MagicMock(spec=import_backend, **kwargs)
        attrs = {'return_value': return_value}
        mock.configure_mock(**attrs)
        return mock

    return importer



from eventsourcing_helpers.repository.snapshot.backends import SnapshotBackend


class NullSnapshotBackend(SnapshotBackend):
    """
    Provide a dummy snapshot backend that does not do anything
    """
    def __init__(self, *args, **kwargs) -> None:
        pass

    def save(self, id: str, data: dict) -> None:
        pass

    def load(self, id: str) -> None:
        return None

    def delete(self, id: str) -> None:
        pass



from typing import Callable

import structlog
from confluent_kafka import KafkaException

from eventsourcing_helpers.metrics import statsd
from eventsourcing_helpers.models import AggregateRoot
from eventsourcing_helpers.repository.snapshot import Snapshot
from eventsourcing_helpers.serializers import from_message_to_dto
from eventsourcing_helpers.utils import import_backend

BACKENDS = {
    'kafka_avro': 'eventsourcing_helpers.repository.backends.kafka.KafkaAvroBackend',
}

logger = structlog.get_logger(__name__)


class Repository:
    """
    Interface to communicate with a repository backend.

    The repository acts as a mediator between the domain and the data mapping
    layer.

    More concrete it provides a way to store and retrieve events that belongs
    to an aggregate root from/to some kind of storage.

    It also handles snapshots by saving/loading the latest state of an
    aggregate root.
    """
    DEFAULT_BACKEND = 'kafka_avro'

    def __init__(
        self, config: dict, aggregate_root_cls: AggregateRoot,
        ignore_missing_apply_methods: bool = False, importer: Callable = import_backend,
        message_deserializer: Callable = from_message_to_dto, snapshot=Snapshot, **kwargs
    ) -> None:
        backend_path = config.get('backend', BACKENDS[self.DEFAULT_BACKEND])
        assert 'backend_config' in config, "You must pass a backend config"
        backend_config = config.get('backend_config')

        logger.info(
            "Using repository backend", backend=backend_path, config=backend_config,
            ignore_missing_apply_methods=ignore_missing_apply_methods
        )
        backend_class = importer(backend_path)

        self.aggregate_root_cls = aggregate_root_cls
        self.message_deserializer = message_deserializer
        self.snapshot = snapshot(config, **kwargs)
        self.backend = backend_class(backend_config, **kwargs)

        self.ignore_missing_apply_methods = ignore_missing_apply_methods

    def commit(self, aggregate_root: AggregateRoot, **kwargs) -> None:
        """
        Commit staged events to the repository.

        Args:
            aggregate_root: Aggregate root with staged events to commit.
        """
        assert isinstance(aggregate_root, AggregateRoot)
        id, events = aggregate_root.id, aggregate_root._events

        if events:
            assert id, "The id must be set on the aggregate root"
            logger.info("Committing staged events to repository")
            self.snapshot.save(aggregate_root)
            try:
                self.backend.commit(id=id, events=events, **kwargs)
            except KafkaException as e:
                logger.info("Kafka commit failed, rolling back snapshot!")
                statsd.increment('eventsourcing_helpers.snapshot.cache.delete', tags=[f'id={id}'])
                self.snapshot.delete(aggregate_root)
                raise e

            aggregate_root._clear_staged_events()

    def load(self, id: str, max_offset: int = None) -> AggregateRoot:
        """
        Load an aggregate root accordingly:
         - First try to load it from the snapshot storage.
         - If there are no snapshot load it from the event storage.

        Args:
            id: ID of the aggregate root.
            max_offset: Stop loading events at this position. If set to None
                all events will be loaded.

        Returns:
            AggregateRoot: Aggregate root instance with the latest state.
        """
        aggregate_root = self._load_from_snapshot_storage(id)
        if aggregate_root is None:
            statsd.increment('eventsourcing_helpers.snapshot.cache.misses')
            aggregate_root = self._load_from_event_storage(id, max_offset)
            logger.debug("Aggregate was loaded from event storage")
        else:
            statsd.increment('eventsourcing_helpers.snapshot.cache.hits')
            logger.debug("Aggregate was loaded from snapshot storage")

        return aggregate_root

    def _load_from_snapshot_storage(self, id: str) -> AggregateRoot:
        aggregate_root = self.snapshot.load(id, self.aggregate_root_cls())

        return aggregate_root

    def _load_from_event_storage(self, id: str, max_offset: int) -> AggregateRoot:
        aggregate_root = self.aggregate_root_cls()
        events = self.backend.get_events(id, max_offset=max_offset)
        events = (self.message_deserializer(event, is_new=False) for event in events)
        aggregate_root._apply_events(
            events, ignore_missing_apply_methods=self.ignore_missing_apply_methods
        )
        return aggregate_root



from copy import deepcopy
from importlib import import_module
from typing import Any, Callable, List


def import_backend(location: str) -> Any:
    """
    Dynamically load a backend.

    Args:
        location: Location of the backend class in the
            backends package. Format: 'module.class'.

    Returns:
        BackendClass: The backend class.

    Example:
        >>> import_backend(
            'eventsourcing_helpers.repository.backends.kafka.KafkaAvroBackend'
        )
        <class 'eventsourcing_helpers.repository.backends.kafka.KafkaAvroBackend'>
    """
    module_name, class_name = location.rsplit('.', 1)
    module = import_module(f'{module_name}')

    backend_cls = getattr(module, class_name)
    return backend_cls


def get_all_nested_keys(data: dict, current_keys: List) -> List:
    """
    Get all keys in a dictoinary. The dictionary could have more levels of
    nested dicts.
    The purpose of this is to detect changes to a class. Therefore the value
    for key: `py/object` is also saved.

    Args:
        data: The dictionary to be examined
        current_keys: The keys found so far (used in recursive calls)

    Returns:
        List: All keys found

    Example:
    >>> get_all_nested_keys(
    ...     {
    ...         '_version': 0,
    ...         'id': None,
    ...         'nested_entity': {
    ...             '__dict__': {
    ...                 'nested_entity_id': 'a'
    ...             },
    ...             'py/object': 'eventsourcing_helpers.models.EntityDict'
    ...         },
    ...         'py/object': 'tests.test_models.NestedAggregate'
    ...     },
    ...     []
    ... )

    [
        '_version', 'id', 'nested_entity', 'py/object', '__dict__',
        'py/object', 'nested_entity_id',
        'eventsourcing_helpers.models.EntityDict',
        'tests.test_models.NestedAggregate'
    ]
    """
    all_keys = deepcopy(current_keys)
    if isinstance(data, dict):
        all_keys.extend(list(data.keys()))
        for key, value in data.items():
            if key == 'py/object':
                all_keys.append(value)
            else:
                all_keys = get_all_nested_keys(value, all_keys)
    elif isinstance(data, (list, tuple)):
        for item in data:
            all_keys = get_all_nested_keys(item, all_keys)

    return all_keys


def get_callable_representation(target: Callable) -> str:
    """
    Get the representation of a callable.

    Ex:
        class Test:
            def method(self):
                pass

            @classmethod
            def klass(cls):
                pass

            @staticmethod
            def static():
                pass

        def myfunc():
            pass

        >> r = get_callable_representation
        >> r(Test), r(Test.method), r(Test.klass), r(Test.static), r(myfunc)
        ('Test', 'Test.method', 'Test.klass', 'Test.static', 'myfunc')

        >> test = Test()
        >> r(test.method), r(test.klass), r(test.static)
        ('Test.method', 'Test.klass', 'Test.static')
    """
    return getattr(target, '__qualname__', getattr(target, '__name__', ''))



import hashlib
from typing import Callable

import structlog

from eventsourcing_helpers.models import AggregateRoot
from eventsourcing_helpers.repository.snapshot.config import get_snapshot_config
from eventsourcing_helpers.repository.snapshot.serializers import (
    from_aggregate_root_to_snapshot, from_snapshot_to_aggregate_root
)
from eventsourcing_helpers.utils import import_backend

BACKENDS = {
    'null': 'eventsourcing_helpers.repository.snapshot.backends.null.NullSnapshotBackend'
}

logger = structlog.get_logger(__name__)


def get_hash(seed: str) -> str:
    """
    Returns a hash of the given seed
    """
    md5_hash = hashlib.md5(seed.encode())
    string_hash = md5_hash.hexdigest()
    return string_hash


class Snapshot:
    """
    Interface to communicate with a snapshot backend.

    The interface provides methods for saving and loading a snapshot
    """
    DEFAULT_BACKEND = 'null'

    def __init__(
        self, config: dict, importer: Callable = import_backend,
        serializer: Callable = from_aggregate_root_to_snapshot,
        deserializer: Callable = from_snapshot_to_aggregate_root,
        hash_function: Callable = get_hash,
        **kwargs
    ) -> None:
        config = get_snapshot_config(config)
        backend_path = config.get('backend', BACKENDS[self.DEFAULT_BACKEND])
        backend_config = config.get('backend_config')

        logger.info(
            "Using snapshot backend", backend=backend_path,
            config=backend_config
        )
        backend_class = importer(backend_path)
        self.backend = backend_class(backend_config, **kwargs)

        self.serializer = serializer
        self.deserializer = deserializer
        self.hash_function = hash_function

    def save(self, aggregate_root: AggregateRoot) -> None:
        """
        Saves an aggregate to the snapshot storage

        Args:
            aggregate_root: The aggregate to be saved
        Returns:
            None
        """
        current_hash = self.hash_function(
            aggregate_root.__class__().get_representation())

        snapshot = self.serializer(aggregate_root, current_hash)
        self.backend.save(aggregate_root.id, snapshot)

    def load(self, id: str, aggregate_root: AggregateRoot) -> AggregateRoot:
        """
        Loads an aggregate root from the snapshot storage.

        If the saved aggregate root hash in the database does not match the
        current hash it means that the model has changed since the snapshot was
        saved.

        If that is the case the snapshot should not be used.

        Args:
            id: ID of the aggregate root.
            Args:
            aggregate_root: The aggregate type of the object
                                            to load

        Returns:
            AggregateRoot: Aggregate root instance with the latest state.
        """
        snapshot = self.backend.load(id)
        current_hash = self.hash_function(
            aggregate_root.__class__().get_representation())
        aggregate_root = self.deserializer(snapshot, current_hash)

        return aggregate_root

    def delete(self, aggregate_root: AggregateRoot) -> None:
        """
        Deletes the snapshot of the aggregate root.

        Args:
            aggregate_root: The aggregate to be deleted

        Returns:
            None
        """
        self.backend.delete(aggregate_root.id)



from copy import deepcopy
from unittest.mock import MagicMock, Mock, patch

import pytest

from eventsourcing_helpers.command_handler import CommandHandler, ESCommandHandler

module = 'eventsourcing_helpers.command_handler'

command_class, id = 'FooCommand', '1'
message = Mock(value={'class': command_class, 'data': {'id': id}})
events = [1, 2, 3]

command = Mock()
command._class = command_class
command.id = id


class ESCommandHandlerTests:
    def setup_method(self):
        self.aggregate_root = Mock()
        self.aggregate_root.foo_method = Mock()

        config = {
            'return_value.load.return_value.__enter__.return_value': events
        }
        self.repository = MagicMock()
        self.repository.configure_mock(**config)

        command_handler = ESCommandHandler
        command_handler.aggregate_root = self.aggregate_root
        command_handler.handlers = {
            command_class: self.aggregate_root.foo_method}
        command_handler.repository_config = {'empty_config': None}

        self.message_deserializer = Mock()
        self.message_deserializer.return_value = command

        self.handler = command_handler(
            message_deserializer=self.message_deserializer,
            repository=self.repository
        )

    @patch(f'{module}.ESCommandHandler._commit_staged_events')
    @patch(f'{module}.ESCommandHandler._handle_command')
    @patch(f'{module}.ESCommandHandler._get_aggregate_root')
    @patch(f'{module}.ESCommandHandler._can_handle_command')
    @patch(f'{module}.statsd.timed')
    def test_handle(self, mock_metrics_timed, mock_can_handle,
                    mock_get, mock_handle, mock_commit
                    ):
        """
        Test that the correct methods are invoked when handling a command.
        """
        mock_metrics_timed.return_value.__enter__.return_value = Mock

        mock_can_handle.return_value = True
        mock_get.return_value = self.aggregate_root

        self.handler.handle(message)

        self.message_deserializer.assert_called_once_with(message)
        mock_can_handle.assert_called_once_with(message)
        mock_get.assert_called_once_with(command.id)
        mock_handle.assert_called_once_with(
            command, handler_inst=self.aggregate_root
        )
        mock_commit.assert_called_once_with(self.aggregate_root)

    @patch(f'{module}.ESCommandHandler._commit_staged_events')
    @patch(f'{module}.ESCommandHandler._handle_command')
    @patch(f'{module}.ESCommandHandler._get_aggregate_root')
    @patch(f'{module}.ESCommandHandler._can_handle_command')
    @patch(f'{module}.statsd.timed')
    def test_handle_deletes_snapshot_on_error(
        self, mock_metrics_timed, mock_can_handle,
        mock_get, mock_handle, mock_commit
    ):
        mock_metrics_timed.return_value.__enter__.return_value = Mock

        mock_can_handle.return_value = True
        mock_get.return_value = self.aggregate_root
        mock_handle.side_effect = TypeError

        with pytest.raises(TypeError):
            self.handler.handle(message)

        self.handler.repository.snapshot.delete.assert_called_once_with(
            self.aggregate_root)

    def test_commit_staged_events(self):
        """
        Test that the repository is invoked correctly.
        """
        self.handler._commit_staged_events(self.aggregate_root)
        self.repository.return_value.commit.assert_called_once_with(
            self.aggregate_root
        )

    def test_handle_command_by_str(self):
        """
        Test that the correct command handler is invoked using a str.
        """
        self.handler._handle_command(command, handler_inst=self.aggregate_root)
        self.aggregate_root.foo_method.called_once_with(command)

    def test_handle_command_by_function(self):
        """
        Test that the correct command handler is invoked using a class function.
        """
        self.handler.handlers = {command_class: self.aggregate_root.foo_method}
        self.handler._handle_command(command, handler_inst=self.aggregate_root)
        self.aggregate_root.foo_method.called_once_with(command)


class CommandHandlerTests:
    def setup_method(self):
        self.mock_handler = Mock()

        command_handler = CommandHandler
        command_handler.handlers = {command_class: self.mock_handler}

        self.message_deserializer = Mock()
        self.message_deserializer.return_value = command

        self.handler = command_handler(
            message_deserializer=self.message_deserializer
        )

    @patch(f'{module}.CommandHandler._handle_command')
    @patch(f'{module}.CommandHandler._can_handle_command')
    @patch(f'{module}.statsd.timed')
    def test_handle(self, mock_metrics_timed, mock_can_handle, mock_handle):
        """
        Test that the correct methods are invoked when handling a command.
        """
        mock_metrics_timed.return_value.__enter__.return_value = Mock

        mock_can_handle.return_value = True
        self.handler.handle(message)

        self.message_deserializer.assert_called_once_with(message)
        mock_can_handle.assert_called_once_with(message)
        mock_handle.assert_called_once_with(command)

    def test_can_handle_command(self):
        """
        Test that we only handle registered commands.
        """
        can_handle = self.handler._can_handle_command(message)
        assert can_handle is True

        _message = deepcopy(message)
        _message.value['class'] = 'BarCommand'
        can_handle = self.handler._can_handle_command(_message)
        assert can_handle is False

    def test_handle_command(self):
        """
        Test that the correct command handler is invoked.
        """
        self.handler._handle_command(command)
        self.mock_handler.called_once_with(command)



import pytest

from eventsourcing_helpers.utils import get_all_nested_keys, get_callable_representation


class Test:
    def method(self):
        pass

    @classmethod
    def klass(cls):
        pass

    @staticmethod
    def static():
        pass


def myfunc():
    pass


@pytest.mark.parametrize('result, expected_result', [
    (get_callable_representation(Test), 'Test'),
    (get_callable_representation(Test()), ''),
    (get_callable_representation(Test().method), 'Test.method'),
    (get_callable_representation(Test.method), 'Test.method'),
    (get_callable_representation(Test().klass), 'Test.klass'),
    (get_callable_representation(Test.klass), 'Test.klass'),
    (get_callable_representation(Test().static), 'Test.static'),
    (get_callable_representation(Test.static), 'Test.static'),
    (get_callable_representation(myfunc), 'myfunc'),
    (get_callable_representation(None), ''),
    (get_callable_representation(1), ''),
])
def test_get_callable_representation(result, expected_result):
    assert expected_result == result


@pytest.mark.parametrize('data, expected_result', [
    (
        {
            '_version': 0,
        },
        [
            '_version',
        ]
    ),
    (
        {
            '_version': 0,
            'id': None,
            'py/object': 'tests.test_models.NestedAggregate'
        },
        [
            '_version',
            'id',
            'py/object',
            'tests.test_models.NestedAggregate',
        ]
    ),
    (
        {
            '_version': 0,
            'id': None,
            'nested_entity': {
                '__dict__': {
                    'nested_entity_id': 'a'
                },
                'py/object': 'eventsourcing_helpers.models.EntityDict'
            },
            'py/object': 'tests.test_models.NestedAggregate'
        },
        [
            '_version',
            'id',
            'nested_entity',
            'py/object',
            'tests.test_models.NestedAggregate',
            'py/object',
            '__dict__',
            'eventsourcing_helpers.models.EntityDict',
            'nested_entity_id'
        ]
    ),
])
def test_get_all_nested_keys(data, expected_result):
    result = get_all_nested_keys(data, [])
    for expected_key in expected_result:
        assert expected_key in result



import structlog
from redis import StrictRedis
from redis.sentinel import Sentinel

from confluent_kafka_helpers.message import Message

from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog.backends import (
    OffsetWatchdogBackend
)

logger = structlog.get_logger(__name__)

__all__ = ['RedisOffsetWatchdogBackend']


class RedisOffsetWatchdogBackend(OffsetWatchdogBackend):
    """
    Redis offset watchdog backend.

    Stores the last offsets in a Redis database.
    """
    DEFAULT_CONFIG = {
        'socket_connect_timeout': 1.0,
        'socket_timeout': 1.0,
        'retry_on_timeout': True,
        'decode_responses': True
    }

    def __init__(self, config: dict) -> None:
        super().__init__(config=config)
        self._redis = None
        self._redis_sentinel = None
        config = {**self.DEFAULT_CONFIG, **config}
        config.pop('group.id', None)

        if 'url' in config:
            assert 'sentinels' not in config
            self._redis = StrictRedis.from_url(**config)
        else:
            assert 'sentinels' in config
            assert 'service_name' in config

            sentinels = [tuple(h.split(':')) for h in config.pop('sentinels').split(',')]
            self._redis_sentinel = Sentinel(sentinels=sentinels, **config)
            self._redis_sentinel_service_name = config.pop('service_name')

    @property
    def redis(self) -> StrictRedis:
        if self._redis:
            return self._redis
        return self._redis_sentinel.master_for(self._redis_sentinel_service_name)  # type: ignore

    def seen(self, message: Message) -> bool:
        last_offset = self.redis.get(self._key(message))
        if last_offset is None:
            return False

        # https://github.com/edenhill/librdkafka/issues/1720
        # until this bug is fixed at least make sure we log when it happens.
        offset = message._meta.offset
        offset_diff = offset - int(last_offset)
        if offset_diff != 1:
            logger.warning("Offset deviation detected", offset_diff=offset_diff)

        return offset <= int(last_offset)

    def set_seen(self, message: Message):
        self.redis.set(self._key(message), message._meta.offset)



import copy
import time
from typing import Union
from unittest.mock import MagicMock

from confluent_kafka_helpers.message import Message


def create_message(message_class: str, data: dict, headers: Union[None, dict]) -> Message:
    kafka_message = MagicMock()
    kafka_message.configure_mock(
        **{
            'value.return_value': {
                'class': message_class,
                'data': copy.deepcopy(data)
            },
            'timestamp.return_value': (0, time.time()),
            'headers.return_value': headers
        }
    )
    message = Message(kafka_message=kafka_message)
    return message



import time
from functools import partial
from typing import Callable

import structlog
from confluent_kafka import KafkaError, KafkaException

from confluent_kafka_helpers.consumer import AvroConsumer
from confluent_kafka_helpers.message import Message
from confluent_kafka_helpers.producer import AvroProducer

from eventsourcing_helpers import metrics
from eventsourcing_helpers.messagebus.backends import MessageBusBackend
from eventsourcing_helpers.messagebus.backends.kafka.config import (
    get_consumer_config, get_offset_watchdog_config, get_producer_config
)
from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog import OffsetWatchdog
from eventsourcing_helpers.serializers import to_message_from_dto

logger = structlog.get_logger(__name__)


class KafkaAvroBackend(MessageBusBackend):
    def __init__(
        self, config: dict, producer: AvroProducer = AvroProducer,
        consumer: AvroConsumer = AvroConsumer, value_serializer: Callable = to_message_from_dto,
        get_producer_config: Callable = get_producer_config,
        get_consumer_config: Callable = get_consumer_config,
        get_offset_watchdog_config: Callable = get_offset_watchdog_config
    ) -> None:
        self.consumer = None
        self.producer = None
        self.offset_watchdog = None

        producer_config = get_producer_config(config)
        consumer_config = get_consumer_config(config)
        offset_wd_config = get_offset_watchdog_config(config)

        if producer_config:
            self.flush = producer_config.pop('flush', False)
            self.producer = producer(producer_config, value_serializer=value_serializer)
        if consumer_config:
            self.consumer = partial(consumer, config=consumer_config)
        if offset_wd_config:
            self.offset_watchdog = OffsetWatchdog(offset_wd_config)

    def _shall_handle(self, message: Message) -> bool:
        if not self.offset_watchdog:
            return True
        return not self.offset_watchdog.seen(message)

    def _set_handled(self, message: Message):
        if self.offset_watchdog:
            try:
                self.offset_watchdog.set_seen(message)
            except Exception:
                logger.exception("Failed to set offset, but will continue")

    @metrics.call_counter('eventsourcing_helpers.messagebus.kafka.handle.count')
    @metrics.timed('eventsourcing_helpers.messagebus.kafka.handle.time')
    def _handle(self, handler: Callable, message: Message, consumer: AvroConsumer) -> None:
        start_time = time.time()
        if self._shall_handle(message):
            handler(message)
            self._set_handled(message)
        if consumer.is_auto_commit is False:
            try:
                consumer.commit(asynchronous=False)
            except KafkaException as e:
                error_code = e.args[0].code()
                if error_code == KafkaError._NO_OFFSET:
                    logger.warning("Offset already committed")
                else:
                    raise
        end_time = time.time() - start_time
        logger.debug(f"Message processed in {end_time:.5f}s")

    def produce(self, value: dict, key: str = None, topic: str = None, **kwargs) -> None:
        assert self.producer is not None, "Producer is not configured"

        while True:
            try:
                self.producer.produce(key=key, value=value, topic=topic, **kwargs)
                break
            except BufferError:
                self.producer.poll(timeout=0.5)
            continue

        self.producer.poll(0)

        if self.flush:
            self.producer.flush()

    def get_consumer(self) -> AvroConsumer:
        assert self.consumer is not None, "Consumer is not configured"
        return self.consumer

    def consume(self, handler: Callable) -> None:
        assert callable(handler), "You must pass a message handler"
        Consumer = self.get_consumer()
        with Consumer() as consumer:
            for message in consumer:
                self._handle(handler, message, consumer)



from typing import Any


class RepositoryBackend:
    """
    Repository interface.
    """

    def commit(self, id: str, events: list, **kwargs) -> None:
        raise NotImplementedError()

    def load(self, id: str, **kwargs) -> Any:
        raise NotADirectoryError()

    def get_events(self, id: str, **kwargs) -> Any:
        raise NotImplementedError()



from typing import NamedTuple
from unittest.mock import patch

from eventsourcing_helpers.message import message_factory
from eventsourcing_helpers.serializers import from_message_to_dto, to_message_from_dto


class Message:
    def __init__(self, data):
        self.value = data
        self._meta = object()


class SerializerTests:

    @patch('eventsourcing_helpers.serializers.message_factory')
    def test_from_message_to_dto(self, mock_factory):
        """
        Test that the message factory is invoked correctly when
        deserializing a message.
        """
        message = Message({'class': 'FooClass', 'data': {'foo': 'bar'}})
        from_message_to_dto(message)

        assert mock_factory.call_args[0][0].__name__ == 'FooClass'
        assert mock_factory.call_args[0][0]._fields == ('foo', 'Meta')

    def test_to_message_from_dto(self):
        """
        Test that we can serialize a DTO to a message.
        """
        fields = [('id', None)]
        FooEvent = message_factory(NamedTuple('FooEvent', fields))
        dto = FooEvent(id=1)
        message = to_message_from_dto(dto)

        assert message['class'] == 'FooEvent'
        assert message['data']['id'] == 1



from confluent_kafka_helpers.message import Message

from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog.backends import (
    OffsetWatchdogBackend
)


class NullOffsetWatchdogBackend(OffsetWatchdogBackend):
    """
    Null (bypass) backend. Does nothing and treats all messages as never seen.
    """
    def seen(self, message: Message) -> bool:
        return False

    def set_seen(self, message: Message):
        pass



from typing import Callable, Iterator

from confluent_kafka_helpers.loader import AvroMessageLoader, MessageGenerator
from confluent_kafka_helpers.message import Message
from confluent_kafka_helpers.producer import AvroProducer

from eventsourcing_helpers.repository.backends import RepositoryBackend
from eventsourcing_helpers.repository.backends.kafka.config import (
    get_loader_config, get_producer_config
)
from eventsourcing_helpers.serializers import to_message_from_dto


class KafkaAvroBackend(RepositoryBackend):
    def __init__(
        self, config: dict, producer=AvroProducer, loader=AvroMessageLoader,
        value_serializer: Callable = to_message_from_dto,
        get_producer_config: Callable = get_producer_config,
        get_loader_config: Callable = get_loader_config
    ) -> None:
        producer_config = get_producer_config(config)
        loader_config = get_loader_config(config)

        self.producer, self.loader = None, None
        if producer_config:
            self.producer = producer(producer_config, value_serializer=value_serializer)
        if loader_config:
            self.loader = loader(loader_config)

    def commit(self, id: str, events: list, **kwargs) -> None:
        """
        Commit staged events.

        Args:
            id: ID of the aggregate root to be used as key in the message.
            events: List of staged events to be committed.
        """
        assert self.producer is not None, "Producer is not configured"

        for event in events:
            self.producer.produce(key=id, value=event, **kwargs)

    def load(self, id: str, **kwargs) -> MessageGenerator:
        """
        Returns the repository message loader.

        Args:
            id: ID of the aggregate root.

        Returns:
            MessageGenerator: Repository message loader.
        """
        assert self.loader is not None, "Loader is not configured"
        return self.loader.load(id, **kwargs)

    def get_events(self, id: str, max_offset: int = None) -> Iterator[Message]:  # type: ignore
        """
        Get all aggregate events from the repository one at a time.

        Args:
            id: ID of the aggregate root.
            max_offset: Stop loading events at this offset.

        Yields:
            Message: The next available event.
        """
        with self.load(id) as events:  # type:ignore
            for event in events:
                if max_offset is not None and event._meta.offset > max_offset:
                    break
                yield event



from typing import Dict, Union

import structlog

logger = structlog.get_logger(__name__)


class SnapshotBackend:
    def save(self, id: str, data: dict) -> None:
        raise NotImplementedError()

    def load(self, id: str) -> Union[Dict, None]:
        raise NotImplementedError()

    def delete(self, id: str) -> None:
        raise NotImplementedError()



from collections import defaultdict
from copy import deepcopy

TOP_LEVEL_PRODUCER_CONFIG_KEYS = (
    'bootstrap.servers',
    'schema.registry.url',
    'security.protocol',
    'sasl.mechanisms',
    'sasl.username',
    'sasl.password',
    'ssl.ca.location',
)

TOP_LEVEL_CONSUMER_CONFIG_KEYS = (
    'bootstrap.servers',
    'schema.registry.url',
    'security.protocol',
    'sasl.mechanisms',
    'sasl.username',
    'sasl.password',
    'ssl.ca.location',
)


def get_producer_config(config):
    config = deepcopy(config)
    producer_config = config.get('producer', None)
    if not producer_config or not producer_config.get('enabled', True):
        return None

    for key in TOP_LEVEL_PRODUCER_CONFIG_KEYS:
        if key in config:
            producer_config.update({f'{key}': config[key]})
    return producer_config


def get_consumer_config(config):
    config = deepcopy(config)
    consumer_config = config.get('consumer')
    if not consumer_config or not consumer_config.get('enabled', True):
        return None

    consumer_config.pop('offset_watchdog', None)
    for key in TOP_LEVEL_CONSUMER_CONFIG_KEYS:
        if key in config:
            consumer_config.update({f'{key}': config[key]})
    return consumer_config


def get_offset_watchdog_config(config):
    config = deepcopy(config)
    consumer_config = config.get('consumer', {})
    if not consumer_config:
        return None

    offset_wd_config = defaultdict(dict, consumer_config.get('offset_watchdog', {}))
    offset_wd_config['backend_config'].update({'group.id': consumer_config['group.id']})
    return offset_wd_config



from unittest.mock import MagicMock

import pytest

from eventsourcing_helpers.repository import Snapshot
from eventsourcing_helpers.repository.backends import RepositoryBackend


@pytest.fixture(scope='function')
def snapshot_mock():
    def snapshot(return_value, **kwargs):
        mock = MagicMock(spec=Snapshot, **kwargs)
        attrs = {'return_value.load.return_value': return_value}
        mock.configure_mock(**attrs)
        return mock

    return snapshot


@pytest.fixture(scope='function')
def repository_backend_mock():
    def repository_backend(return_value, **kwargs):
        mock = MagicMock(spec=RepositoryBackend, **kwargs)
        attrs = {
            'return_value.load.return_value.__enter__.return_value': return_value,
            'return_value.get_events.return_value': return_value
        }
        mock.configure_mock(**attrs)
        return mock

    return repository_backend



"""
In certain cases, messages from Kafka are consumed more than once.
To get as possible to "exaclty once" semantics, we implement our
own Offset Watchdog that provides the best effort of controlling
"double reads" by storing the last offset for the particular
(topic, partition, consumer group) tuple and skipping messages
with offset lower than the stored one.

Certain backend implementations are available:
 * memory (default) - stores offsets in a map, thus works only
   for the current process and not restart-safe;
 * redis - stores offsets in a Redis instance specified via REDIS_URL;
 * null - bypasses all the checks (for using in tests, debugging, etc.)
"""
from typing import Callable

import structlog

from confluent_kafka_helpers.message import Message

from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog.backends import (
    OffsetWatchdogBackend
)
from eventsourcing_helpers.metrics import base_metric, statsd
from eventsourcing_helpers.utils import import_backend

BACKENDS_PATH = 'eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog.backends'
BACKENDS = {
    'null': f'{BACKENDS_PATH}.null.NullOffsetWatchdogBackend',
    'memory': f'{BACKENDS_PATH}.memory.InMemoryOffsetWatchdogBackend',
    'redis': f'{BACKENDS_PATH}.redis.RedisOffsetWatchdogBackend'
}

logger = structlog.get_logger(__name__)


class OffsetWatchdog:
    """
    Offset watchdog facade.

    Loads and configures the real storage backend on initialization.
    """
    DEFAULT_BACKEND = 'memory'

    def __init__(self, config: dict, importer: Callable = import_backend) -> None:
        backend_path = config.get('backend', BACKENDS[self.DEFAULT_BACKEND])
        backend_config = config.get('backend_config')
        self.config = backend_config

        logger.debug("Using offset watchdog backend", backend=backend_path, config=self.config)
        backend_class = importer(backend_path)
        self.backend: OffsetWatchdogBackend = backend_class(config=self.config)

    def seen(self, message: Message) -> bool:
        """Checks if the `message` has been seen before"""
        seen = self.backend.seen(message)
        if seen:
            logger.warning("Message already seen previously", message_meta=message._meta)
            statsd.increment(
                f'{base_metric}.messagebus.kafka.offset_watchdog.seen.count', tags=[
                    f'partition:{message._meta.partition}', f'topic:{message._meta.topic}',
                    f'consumer_group:{self.config["group.id"]}'  # type: ignore
                ]
            )
        return seen

    def set_seen(self, message: Message):
        """Marks the message as already seen"""
        self.backend.set_seen(message)



from functools import wraps

base_metric = 'eventsourcing_helpers'


class StatsdNullClient:
    """
    No-op datadog statsd client implementing the null object pattern.
    """
    __call__ = __getattr__ = lambda self, *_, **__: self

    def timed(self, *args, **kwargs):
        return TimedNullDecorator()


class TimedNullDecorator:
    __enter__ = __getattr__ = lambda self, *_, **__: self

    def __call__(self, f):
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapped

    def __exit__(self, *args):
        pass


try:
    import datadog
    statsd = datadog.statsd
except ModuleNotFoundError:
    statsd = StatsdNullClient()


def call_counter(base_metric):
    def wrapped(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            statsd.increment(f"{base_metric}.total")
            try:
                return f(*args, **kwargs)
            except Exception:
                statsd.increment(f"{base_metric}.error")
                raise
        return decorator
    return wrapped


def timed(base_metric, tags=None):
    if tags is None:
        tags = []

    def wrapped(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            with statsd.timed(f"{base_metric}.time", tags=tags):
                return f(*args, **kwargs)
        return decorator
    return wrapped



from functools import partial
from unittest.mock import Mock

import pytest
from confluent_kafka import KafkaException

from eventsourcing_helpers.repository import Repository


class RepositoryTests:
    aggregate_events = [1, 2, 3]
    config = {'backend_config': {}}

    @pytest.fixture(autouse=True)
    def setup_method(
        self, aggregate_root_cls_mock, snapshot_mock, repository_backend_mock,
        importer_mock
    ):
        self.aggregate_root_cls = aggregate_root_cls_mock()
        self.snapshot = snapshot_mock(return_value=None)
        self.repository_backend = repository_backend_mock(
            return_value=self.aggregate_events
        )
        self.importer = importer_mock(return_value=self.repository_backend)
        self.message_deserializer = lambda e, *args, **kwargs: e

        self.repository = partial(
            Repository, config=self.config,
            aggregate_root_cls=self.aggregate_root_cls, importer=self.importer,
            message_deserializer=self.message_deserializer,
            snapshot=self.snapshot
        )

    def test_should_load_aggr_root_from_event_storage(self):
        repository = self.repository()
        aggregate_root = repository.load(id=1)

        assert repository.snapshot.load.called is True
        assert repository.snapshot.load.return_value is None
        assert repository.backend.get_events.called is True
        assert aggregate_root is not None

    def test_should_load_aggr_root_from_snapshot_storage(self, snapshot_mock):
        snapshot = snapshot_mock(return_value=Mock())
        repository = self.repository(snapshot=snapshot)
        aggregate_root = repository.load(id=1)

        assert repository.snapshot.load.called is True
        assert repository.backend.get_events.called is False
        assert aggregate_root is not None

    def test_should_apply_events_when_loading_from_event_storage(
        self, aggregate_root_cls_mock
    ):
        aggregate_root_cls = aggregate_root_cls_mock(exhaust_events=False)
        repository = self.repository(aggregate_root_cls=aggregate_root_cls)
        aggregate_root = repository.load(id=1)

        args, kwargs = aggregate_root._apply_events.call_args
        events, *_ = args

        assert aggregate_root._apply_events.called is True
        assert list(events) == self.aggregate_events

    def test_repository_commit_should_call_backend_and_snapshot(
        self, aggregate_root_cls_mock
    ):
        aggregate_root_cls = aggregate_root_cls_mock(exhaust_events=False)
        aggregate_root_cls.id = 1

        repository = self.repository(aggregate_root_cls=aggregate_root_cls)
        repository.commit(aggregate_root_cls)

        assert repository.backend.commit.called is True
        assert repository.snapshot.save.called is True

    def test_repository_commit_should_delete_snapshot_on_kafka_exception(
        self, aggregate_root_cls_mock
    ):
        aggregate_root_cls = aggregate_root_cls_mock(exhaust_events=False)
        aggregate_root_cls.id = 1

        repository = self.repository(aggregate_root_cls=aggregate_root_cls)
        repository.backend.commit.side_effect = KafkaException

        with pytest.raises(KafkaException):
            repository.commit(aggregate_root_cls)

            assert repository.snapshot.save.called is True
            assert repository.snapshot.delete.called is True



from eventsourcing_helpers.handler import Handler
from eventsourcing_helpers.messagebus import MessageBus


class Consumer:
    """
    Helper to consume and handle messages from the bus.
    """
    def __init__(self, messagebus: MessageBus, handler: Handler) -> None:
        self._messagebus = messagebus
        self._handler = handler

    def consume(self) -> None:
        self._messagebus.consume(handler=self._handler.handle)



from unittest.mock import Mock

from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog import OffsetWatchdog
from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog.backends.memory import (
    InMemoryOffsetWatchdogBackend
)
from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog.backends.null import (
    NullOffsetWatchdogBackend
)


class OffsetWatchdogTests:
    def setup_method(self):
        self.old_message = self._create_message(8)
        self.current_message = self._create_message(88)
        self.future_message = self._create_message(888)

    def _create_message(self, offset: int):
        message = Mock()
        message._meta = Mock()
        message._meta.offset = offset
        message._meta.partition = 1
        message._meta.topic = "some-topic"
        return message

    def test_nulloffsetwatchdogbackend(self):
        backend = NullOffsetWatchdogBackend(config={'group.id': 'consumer'})
        backend.set_seen(self.current_message)
        for message in (self.old_message, self.current_message,
                        self.future_message):
            assert not backend.seen(message)

    def test_inmemoryoffsetwatchdogbackend(self):
        backend = InMemoryOffsetWatchdogBackend(config={'group.id': 'consumer'})
        backend.set_seen(self.current_message)
        assert backend.seen(self.old_message)
        assert backend.seen(self.current_message)
        assert not backend.seen(self.future_message)

    def test_default_backend_configured(self):
        watchdog = OffsetWatchdog(
            config={
                'backend_config': {'group.id': 'consumer'}
            }
        )
        assert watchdog.backend



from collections import deque
from dataclasses import dataclass, field
from typing import Callable, Deque

import structlog

from confluent_kafka_helpers.message import Message

from eventsourcing_helpers.messagebus.backends import MessageBusBackend
from eventsourcing_helpers.messagebus.backends.mock.utils import create_message

logger = structlog.get_logger(__name__)


@dataclass
class Consumer:
    messages: Deque[Message] = field(default_factory=deque)

    def add_message(self, message_class: str, data: dict, headers: dict = None) -> None:
        message = create_message(message_class=message_class, data=data, headers=headers)
        self.messages.append(message)

    def get_messages(self) -> Deque[Message]:
        return self.messages

    def assert_one_message_added_with(
        self, message_class: str, data: dict, headers: dict = None
    ) -> None:
        if headers is None:
            headers = {}
        assert len(self.messages) == 1
        assert {'class': message_class, 'data': data} == self.messages[0].value
        assert headers == self.messages[0]._meta.headers


@dataclass
class Producer:
    messages: Deque[dict] = field(default_factory=deque)

    def add_message(self, message: dict) -> None:
        self.messages.append(message)

    def clear_messages(self) -> None:
        self.messages.clear()

    def assert_one_message_produced_with(self, key: str, value: dict, **kwargs) -> None:
        assert len(self.messages) == 1
        assert dict(key=key, value=value, **kwargs) in self.messages

    def assert_no_messages_produced(self) -> None:
        assert len(self.messages) == 0


class MockBackend(MessageBusBackend):
    def __init__(self, config: dict) -> None:
        self.consumer = Consumer()
        self.producer = Producer()

    def produce(self, value: dict, key: str = None, **kwargs) -> None:
        self.producer.add_message(dict(value=value, key=key, **kwargs))

    def consume(self, handler: Callable) -> None:
        messages = self.consumer.get_messages()
        while messages:
            handler(messages.popleft())



from confluent_kafka_helpers.message import Message as ConfluentKafkaMessage

from eventsourcing_helpers.message import Message, message_factory

try:
    from cnamedtuple import namedtuple
except ImportError:
    from collections import namedtuple


def from_message_to_dto(message: ConfluentKafkaMessage, is_new=True) -> Message:
    """
    Deserialize a `confluent_kafka_helpers.message.Message` to a data transfer
    object (DTO).

    The message is a dict expected to include two keys `class` and `data`.

    The `class` defines the type of the DTO and `data` the attributes.

    Args:
        message: Message to deserialize.
        is_new: Flag that indicates if the message is new or loaded
            from the repository.

    Returns:
        Message: DTO instance of type 'class' hydrated with 'data'.

    Example:
        >>> message = {
        ...    'class': 'OrderCompleted',
        ...    'data': {
        ...        'order_id': 'UA123',
        ...        'date': '2017-09-08'
        ...    }
        ... }
        >>> from_message_to_dto(message)
        OrderCompleted(order_id='UA123', date='2017-09-08')
    """
    data, class_name = message.value['data'], message.value['class']
    data['Meta'] = message._meta

    message_cls = namedtuple(class_name, data.keys())
    dto = message_factory(message_cls, is_new=is_new)(**data)

    return dto


def to_message_from_dto(dto: Message) -> dict:
    """
    Serialize a data transfer object (DTO) to a message.

    The message includes two keys `class` and `data`. The `class` will be the
    type of the DTO and the `data` will be a dict with all attributes.

    Args:
        dto: DTO instance.

    Returns:
        dict: Serialized message.

    Example:
        >>> OrderCompleted = namedtuple('OrderCompleted', ['order_id', 'date'])
        >>> dto = OrderCompleted('UA123', '2017-09-08')
        >>> to_message_from_dto(dto)
        {
            'class': 'OrderCompleted',
            'data': {
                'order_id': 'UA123',
                'date': '2017-09-08'
            }
        }
    """
    message = {'class': dto._class, 'data': dto.to_dict()}

    return message



from typing import Callable

from eventsourcing_helpers.serializers import from_message_to_dto


class Handler:
    handlers: dict = {}

    def __init__(
        self, message_deserializer: Callable = from_message_to_dto
    ) -> None:
        assert self.handlers
        self.message_deserializer = message_deserializer

    def handle(self, message: dict) -> None:
        raise NotImplementedError(
            "You need to implement handle method in your subclass"
        )



from typing import NamedTuple

import pytest

from eventsourcing_helpers.message import Message, message_factory


class MessageTests:

    def setup_method(self):
        self.data = {'id': 1, 'foo': 'bar', 'baz': None, 'foobar': {'a': 'b'}}
        fields = [(k, None) for k in self.data.keys()]
        self.namedtuple = NamedTuple('FooEvent', fields)
        message_cls = message_factory(self.namedtuple, is_new=True)
        self.message = message_cls(**self.data)

    def test_type(self):
        """
        Test that the message has the correct types.
        """
        assert isinstance(self.message, Message)
        assert type(self.message).__name__ == 'FooEvent'

    def test_message_attr_redirect(self):
        """
        Test that attribute lookups are redirected to the namedtuple.
        """
        assert self.message.id == self.data['id']
        assert self.message.foo == self.data['foo']

    def test_access_missing_attr_on_new_msg_should_raise_attribute_error(self):
        with pytest.raises(AttributeError):
            self.message.boo

    def test_access_missing_attr_on_old_msg_should_return_none(self):
        message_cls = message_factory(self.namedtuple, is_new=False)
        message = message_cls(**self.data)
        assert message.boo is None

    def test_to_dict(self):
        """
        Test that the message is serialized correctly.
        """
        assert self.message.to_dict() == self.data

    def test_name(self):
        """
        Test that the correct name is returned.
        """
        assert self.message._class == self.namedtuple.__name__

    def test_read_only(self):
        """
        Test that we can't mutate the state of a message.
        """
        with pytest.raises(AttributeError):
            self.message.id = 2

    def test_read_only_nested_data_type_new_message(self):
        foobar = self.message.foobar
        foobar['a'] = 'c'
        assert self.message.foobar['a'] == 'b'

    def test_read_only_nested_data_type_old_message(self):
        message_cls = message_factory(self.namedtuple, is_new=False)
        self.message = message_cls(**self.data)

        foobar = self.message.foobar
        foobar['a'] = 'c'
        assert self.message.foobar['a'] == 'b'



import structlog

from confluent_kafka_helpers.message import Message

logger = structlog.get_logger(__name__)


class OffsetWatchdogBackend:
    """
    Abstract base class for the offset watchdog backends
    """

    def __init__(self, config: dict) -> None:
        self.config = config

    def _key(self, message: Message):
        """Returns the key for storing offset in the backend"""
        return f'{message._meta.partition}-{message._meta.topic}-{self.config["group.id"]}'

    def seen(self, message: Message) -> bool:
        """Checks if the `message` has been seen before"""
        raise NotImplementedError

    def set_seen(self, message: Message):
        """Marks the message as already seen"""
        raise NotImplementedError



from typing import Callable, Union

import jsonpickle

from eventsourcing_helpers.models import AggregateRoot


def from_aggregate_root_to_snapshot(
    aggregate_root: AggregateRoot,
    current_hash: int,
    encoder: Callable = jsonpickle.encode
) -> dict:
    """
    Serializes an aggregate root into a format suitable for snapshot storage

    Args:
        aggregate_root: The aggregate root to be saved
        current_hash: The hash of the aggregate_root schema
        encoder (optional): The function that encodes the aggregate
            root into data for storage

    Returns:
        dict: The data blob to be stored in the snapshot storage

    Example:
    >>> from_aggregate_root_to_snapshot(aggregate, 123456)
    {
        'data': '{"py/object": "eventsourcing_helpers.models.AggregateRoot",
        "_version": 0, "id": null}', 'version': 0, 'hash': 123456
    }

    """
    snapshot = {
        'data': encoder(aggregate_root),
        'version': aggregate_root._version,
        'hash': current_hash
    }
    return snapshot


def from_snapshot_to_aggregate_root(
    snapshot: dict, current_hash: int, decoder: Callable = jsonpickle.decode
) -> Union[AggregateRoot, None]:
    """Converts the snapshot data into an AggregateRoot (or child)

    Args:
        snapshot: the snapshot data
        current_hash: The current hash of the schema of the object to be
            created (AggregateRoot or child)
        decoder (optional): The decoder of the snapshot data

    Returns:
        AggregateRoot: The restored object

    Example:
    >>> from_snapshot_to_aggregate_root(snapshot, 123456)
    AggregateRoot({'_version': 0})
    """

    if not snapshot:
        return None

    data, hash = snapshot['data'], snapshot['hash']
    if data and current_hash == hash:
        return decoder(data)
    else:
        return None



from unittest.mock import MagicMock

from eventsourcing_helpers.models import AggregateRoot
from eventsourcing_helpers.repository.builder import AggregateBuilder


class TestAggregate(AggregateRoot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.state = None

    def apply_first_event(self, event):
        self.state = "first_event"

    def apply_second_event(self, event):
        self.state = "second_event"

    def apply_third_event(self, event):
        self.state = "third_event"


class AggregateBuilderTests:
    def setup_method(self):
        events = [
            MagicMock(id="123", _class="FirstEvent", _meta=MagicMock(offset=0)),
            MagicMock(id="123", _class="SecondEvent", _meta=MagicMock(offset=1)),
            MagicMock(id="123", _class="ThirdEvent", _meta=MagicMock(offset=2))
        ]
        loader = MagicMock()
        loader.configure_mock(**{
            'return_value.load.return_value.__enter__.return_value': events
        })
        config = {
            'backend_config': {'loader': {'foo': 'bar'}}
        }
        self.factory = AggregateBuilder(
            config=config, aggregate_root_cls=TestAggregate,
            message_deserializer=lambda e, **kwargs: e, loader=loader
        )

    def test_rebuild_returns_the_aggregate_to_the_final_state(self):
        aggregate = self.factory.rebuild(id="123")
        assert aggregate.state == "third_event"

    def test_rebuild_returns_the_aggregate_in_the_right_state_when_max_offset_is_given(self):
        aggregate = self.factory.rebuild(id="123", max_offset=0)
        assert aggregate.state == "first_event"

        aggregate = self.factory.rebuild(id="123", max_offset=1)
        assert aggregate.state == "second_event"

        aggregate = self.factory.rebuild(id="123", max_offset=2)
        assert aggregate.state == "third_event"



from typing import Type

from eventsourcing_helpers.models import AggregateRoot
from eventsourcing_helpers.repository import Repository


class AggregateBuilder:
    def __init__(
        self, config, aggregate_root_cls: AggregateRoot, repository: Type[Repository] = Repository,
        **kwargs
    ) -> None:
        self.repository = repository(
            config, aggregate_root_cls=aggregate_root_cls, ignore_missing_apply_methods=True,
            **kwargs
        )

    def rebuild(self, id: str, max_offset: int = None) -> AggregateRoot:
        """
        Rebuild aggregate from the repository.

        Max offset is used to make sure we load the aggregate up until the point
        where the event was created.

        Args:
            id: ID of the aggregate root to load.
            max_offset: Stop loading events at this position.
        """
        aggregate = self.repository.load(id, max_offset=max_offset)

        return aggregate



from unittest.mock import call, patch

import pytest

from eventsourcing_helpers.metrics import StatsdNullClient, call_counter


class StatsdNullClientTests:
    def setup_method(self):
        self.client = StatsdNullClient()

    def test_method_calls_should_not_fail(self):
        self.client.increment('foo', 1)
        self.client.foo('baz', 2)

    def test_timed_decorator_should_execute_decorated_method(self):
        @self.client.timed('foo', 1)
        def foo(a):
            return a

        value = foo('bar')
        assert value == 'bar'


class CallCounterTests:
    @patch('eventsourcing_helpers.metrics.statsd')
    def test_increments_total_count(self, mock_statsd):
        @call_counter('foo.count')
        def foo(a):
            pass

        foo('a')
        mock_statsd.increment.assert_called_once_with('foo.count.total')

    @patch('eventsourcing_helpers.metrics.statsd')
    def test_increments_counts_and_reraises_exception(self, mock_statsd):
        @call_counter('foo.count')
        def foo(a):
            1 / 0

        with pytest.raises(ZeroDivisionError):
            foo('a')
        expected_calls = [call('foo.count.total'), call('foo.count.error')]
        mock_statsd.increment.assert_has_calls(expected_calls)



from typing import Dict, Union

from pymongo import MongoClient
from pymongo.errors import PyMongoError

from eventsourcing_helpers.repository.snapshot.backends import SnapshotBackend

DEFAULT_CONFIG = {
    'connectTimeoutMS': 2000,
    'serverSelectionTimeoutMS': 1000,
    'waitQueueTimeoutMS': 1000,
    'socketTimeoutMS': 1000,
}


class MongoSnapshotBackend(SnapshotBackend):
    def __init__(self, config: dict, mongo_client_class=MongoClient) -> None:
        assert 'host' in config, 'You must specify host!'
        mongo_config = {**DEFAULT_CONFIG, **config}
        self.client = mongo_client_class(**mongo_config)
        self.db = self.client.snapshots

    def save(self, id: str, data: dict) -> None:
        """
        Saves the data to the snapshot storage

        Args:
            id (str): The id to save the data for
            data (dict): data to be saved
        Returns:
            None
        """
        query = {'_id': id}
        self.db.snapshots.find_one_and_replace(query, data, upsert=True)

    def load(self, id: str) -> Union[Dict, None]:
        """
        Get the aggregate with the specific id from the snapshot storage

        Args:
            id (str): The id to retrieve the data for
        Returns:
            dict: The stored snapshot data
        """
        query = {'_id': id}
        try:
            snapshot_data = self.db.snapshots.find_one(query)
            return snapshot_data
        except PyMongoError:
            return None

    def delete(self, id: str) -> None:
        """
        Deletes the data of the snapshot with specified id.

        Args:
            id (str): The id to save the data for
        Returns:
            None
        """
        query = {'_id': id}
        self.db.snapshots.delete_many(query)



from confluent_kafka_helpers.message import Message

from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog.backends import (
    OffsetWatchdogBackend
)


class InMemoryOffsetWatchdogBackend(OffsetWatchdogBackend):
    """
    In-memory offset watchdog backend.
    Stores the last offsets in a instance's dictionary
    """
    def __init__(self, config: dict) -> None:
        super().__init__(config=config)
        self._offset_map: dict = {}

    def seen(self, message: Message) -> bool:
        last_offset = self._offset_map.get(self._key(message), -1)
        return message._meta.offset <= last_offset

    def set_seen(self, message: Message):
        self._offset_map[self._key(message)] = message._meta.offset



from setuptools import find_packages, setup

setup(
    name="eventsourcing-helpers",
    version="0.8.8",
    description="Helpers for practicing the Event sourcing pattern",
    url="https://github.com/fyndiq/eventsourcing_helpers",
    author="Fyndiq AB",
    author_email="support@fyndiq.com",
    license="MIT",
    packages=find_packages(),
    setup_requires=['wheel'],
    install_requires=[
        'structlog>=17.2.0',
        'jsonpickle==0.9.6',
        'confluent-kafka-helpers==0.7.5'
    ],
    extras_require={
        'mongo': ["pymongo==3.6.1"],
        'redis': ["redis>=2.10.6", "hiredis>=0.2.0"],
        'cnamedtuple': ["cnamedtuple>=0.1.6"]
    },
    zip_safe=False
)



from functools import partial
from unittest.mock import MagicMock, Mock

from eventsourcing_helpers.messagebus.backends.kafka import KafkaAvroBackend


class AvroConsumerMock:
    def __init__(self, config):
        message = Mock()
        message._meta = Mock()
        message._meta.offset = 1
        self.messages = iter([message, message])

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.messages)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass

    is_auto_commit = False
    commit = Mock()


class KafkaBackendTests:

    def setup_method(self):
        self.consumer = AvroConsumerMock
        self.producer = Mock()
        self.value_serializer = Mock()
        self.config = {'producer': {'foo': 'bar'}, 'consumer': {'group.id': 'consumer.group.1'}}
        self.get_producer_config = Mock(return_value=self.config['producer'])
        self.get_consumer_config = Mock(return_value=self.config['consumer'])
        self.id, self.event = 1, {'event': 'value'}
        self.backend = partial(
            KafkaAvroBackend, self.config, self.producer, self.consumer,
            self.value_serializer, self.get_producer_config, self.get_consumer_config
        )

    def test_init(self):
        """
        Test that the dependencies are setup correctly when
        initializing the backend.
        """
        self.backend()

        self.producer.assert_called_once_with(self.config['producer'], value_serializer=self.value_serializer)

    def test_consume(self):
        backend = self.backend()
        handler = Mock()
        backend.consume(handler=handler)
        assert handler.call_count == 1



from typing import Callable

import structlog

from eventsourcing_helpers.utils import import_backend

BACKENDS = {
    'kafka_avro': 'eventsourcing_helpers.messagebus.backends.kafka.KafkaAvroBackend',
    'mock': 'eventsourcing_helpers.messagebus.backends.mock.MockBackend'
}

logger = structlog.get_logger(__name__)


class MessageBus:
    """
    Interface to communicate with a message bus backend.

    The interface provides methods for producing and consuming messages
    (commands, events).
    """
    DEFAULT_BACKEND = 'kafka_avro'

    def __init__(self, config: dict, importer: Callable = import_backend, **kwargs) -> None:
        backend_path = config.get('backend', BACKENDS[self.DEFAULT_BACKEND])
        assert 'backend_config' in config, "You must pass a backend config"
        backend_config = config.get('backend_config')

        logger.debug("Using message bus backend", backend=backend_path, config=backend_config)
        backend_class = importer(backend_path)
        self.backend = backend_class(config=backend_config, **kwargs)

    def produce(self, value, key=None, **kwargs):
        """
        Produce a message to the message bus.

        Args:
            key: Key to produce message with.
            value: Value to produce message with.
        """
        self.backend.produce(key=key, value=value, **kwargs)

    def get_consumer(self):
        """
        Get a message consumer.

        Use this if you want more control of the message loop.

        Returns:
            Callable: Message consumer.
        """
        return self.backend.get_consumer()

    def consume(self, handler: Callable):
        """
        Consume and handle messages indefinitely.
        """
        self.backend.consume(handler)



from unittest.mock import Mock, patch

import pytest

from eventsourcing_helpers.models import AggregateRoot, Entity, EntityDict


class Foo(AggregateRoot):
    pass


class Bar(Entity):
    pass


class LargerTestAggregate(AggregateRoot):
    def __init__(self):
        super().__init__()
        self.title = None
        self.description = None
        self.status = None
        self.tags = None
        self.properties = None


class NestedAggregate(AggregateRoot):
    def __init__(self):
        super().__init__()
        self.nested_entity = EntityDict()
        self.nested_entity.nested_entity_id = 'a'


class DoubleNestedAggregate(AggregateRoot):
    def __init__(self):
        super().__init__()
        self.nested_entity = EntityDict()
        self.nested_entity.nested_entity_id = 'a'
        self.nested_entity.second_nested_entity = EntityDict()
        self.nested_entity.second_nested_entity.second_nested_entity_id = 'a'


class AggregateRootTests:

    def test_subclass(self):
        """
        Test that the aggregate root is just a subclass of `Entity`.
        """
        aggregate_root = Foo()

        assert isinstance(aggregate_root, AggregateRoot)
        assert isinstance(aggregate_root, Entity)


class EntityTests:

    def setup_method(self):
        self.aggregate_root = Foo()
        self.entity = Bar()
        self.event = Mock()
        self.event._class = 'FooEvent'
        self.event.id = 1

    @patch('eventsourcing_helpers.models.Entity._get_apply_method_name')
    @patch('eventsourcing_helpers.models.Entity._get_entity')
    @patch('eventsourcing_helpers.models.Entity._apply_event')
    def test_apply_event(self, mock_event, mock_entity, mock_apply):
        """
        Test that correct methods are invoked when applying an event.
        """
        mock_apply.return_value = 'apply_foo_event'
        is_new = False

        self.aggregate_root.apply_event(self.event)

        mock_apply.called_once_with(self.event.__class__.__name__)
        mock_entity.called_once_with(self.event.id)
        mock_event.called_once_with(
            self.event, self.aggregate_root, 'apply_foo_event', is_new
        )

    @patch('eventsourcing_helpers.models.Entity.apply_event')
    def test_apply_events(self, mock_apply):
        """
        Test that the apply method is run multiple times.
        """
        num_events = 4
        events = [None] * num_events

        self.aggregate_root._apply_events(events)
        assert mock_apply.call_count == num_events

    def test_clear_events(self):
        """
        Test that the committed events are cleared from ALL
        `Entity` instances.
        """
        self.aggregate_root._events.append(self.event)
        self.entity._events.append(self.event)

        assert len(self.aggregate_root._events) == 2
        self.aggregate_root._clear_staged_events()

        assert len(self.aggregate_root._events) == 0
        assert len(self.entity._events) == 0

    def test_create_id(self):
        """
        Test that the returned id is a string and with correct length.
        """
        id = self.entity.create_id()

        assert isinstance(id, str)
        assert len(id) == 36

    def test_get_apply_method_name(self):
        """
        Test that we get the correct apply method name for an event.
        """
        apply_method_name = self.entity._get_apply_method_name(self.event._class)
        assert apply_method_name == 'apply_foo_event'

    @patch('eventsourcing_helpers.models.Entity._get_all_entities')
    def test_get_entity(self, mock_entities):
        """
        Test that the correct entity is returned.
        """
        self.aggregate_root.id = 1
        self.entity.id = 2

        entities = [self.aggregate_root, self.entity]
        mock_entities.return_value = entities
        entity = self.aggregate_root._get_entity(self.entity.id)
        assert entity == self.entity

        entities.pop()
        mock_entities.return_value = entities
        entity = self.aggregate_root._get_entity(self.entity.id)
        assert entity == self.aggregate_root

    @patch('eventsourcing_helpers.models.Entity._get_child_entities')
    def test_get_all_entities(self, mock_entities):
        """
        Test that we get all child entities including the current instance.
        """
        mock_entities.return_value = [self.entity]
        entities = self.aggregate_root._get_all_entities()

        assert list(entities) == [self.aggregate_root, self.entity]

    @patch('eventsourcing_helpers.models.Entity._get_all_entities')
    def test_get_child_entities(self, mock_entities):
        """
        Test that we get all child entities for the current instance.
        """
        self.aggregate_root.__dict__.update({'Bar': self.entity})
        mock_entities.return_value = [self.entity]
        entities = self.aggregate_root._get_child_entities()

        assert list(entities) == [self.entity]

    def test_stage_event(self):
        """
        Test that an event is staged correctly.
        """
        self.entity._stage_event(self.event, is_new=False)
        assert len(self.entity._events) == 0

        self.entity._stage_event(self.event, is_new=True)
        assert len(self.entity._events) == 1
        assert self.event in self.entity._events

    def test_get_apply_method(self):
        """
        Test that the correct method is returned.
        """
        mock_method = Mock()
        self.aggregate_root.apply_foo_event = mock_method
        method = self.aggregate_root._get_apply_method(
            self.aggregate_root, 'apply_foo_event'
        )
        assert method == mock_method

    @patch('eventsourcing_helpers.models.Entity._get_apply_method')
    @patch('eventsourcing_helpers.models.Entity._stage_event')
    def test_apply_event_aggregate_root(self, mock_stage, mock_apply):
        """
        Test that an event is correctly applied on the aggregate root.
        """
        method_name, is_new = 'apply_foo_event', True

        self.aggregate_root._apply_event(
            self.event, self.aggregate_root, method_name, is_new
        )
        mock_apply.assert_called_once_with(self.aggregate_root, method_name)
        mock_stage.assert_called_once_with(self.event, is_new)

    @pytest.mark.parametrize('entity, data', [
        (Foo(), ['Foo', 'id', '_version']),
        (LargerTestAggregate(), [
            'LargerTestAggregate', 'id', '_version', 'title', 'description',
            'status', 'tags', 'properties'
        ]),
    ])
    def test_get_representation_includes_name_and_fields(self, entity, data):
        representation = entity.get_representation()
        for field in data:
            assert field in representation

    @pytest.mark.parametrize('entity, data', [
        (NestedAggregate(), ['NestedAggregate',
                             'id', '_version',
                             'nested_entity', 'nested_entity_id',
                             'eventsourcing_helpers.models.EntityDict']),
        (DoubleNestedAggregate(), [
            'DoubleNestedAggregate', 'id', '_version',
            'nested_entity', 'nested_entity_id',
            'second_nested_entity', 'second_nested_entity_id',
            'eventsourcing_helpers.models.EntityDict']),
    ])
    def test_get_representation_handles_nested_entities(
        self, entity, data
    ):
        representation = entity.get_representation()
        for field in data:
            assert field in representation


class EntityDictTests:

    def setup_method(self):
        self.entity = Bar()
        self.entity_dict = EntityDict({1: self.entity})

    def test_values(self):
        """
        Test that the value must be of instance `Entity`.
        """
        with pytest.raises(AssertionError):
            self.entity_dict[1] = {1: 'Foo'}

        self.entity_dict[2] = Foo()

    @patch('eventsourcing_helpers.models.EntityDict._get_child_entities')
    def test_get_all_entities(self, mock_entities):
        """
        Test that correct methods are invoked.
        """
        self.entity_dict._get_all_entities()
        mock_entities.assert_called_once()

    @patch('eventsourcing_helpers.models.Entity._get_all_entities')
    def test_get_child_entities(self, mock_entities):
        """
        Test that we get all entities in the current EntityDict.
        """
        mock_entities.return_value = [self.entity]
        entities = self.entity_dict._get_child_entities()

        assert list(entities) == [self.entity]
        mock_entities.assert_called_once()
